random_seed = os.time()
math.randomseed( random_seed )
Log("random seed: "..random_seed)
local pseudorandom = math.random(1, 100)

function random( what )
	return what[ math.random(1, #what) ]
end

function deep_copy( src )
	local ret = {}
	for key, value in pairs( src ) do
		if type(value) == 'table' then
			ret[key] = deep_copy( value )
		else
			ret[key] = value
		end
	end
	return ret
end

function RGB( r, g, b )
	return { r/255, g/255, b/255, 1 }
end

function deep_concat( tab, sep )
	local st = {}
	if #tab > 0 then
		for i=1, #tab do
			if type( tab[i] ) == 'table' then
				table.insert( st, "{" )
				table.insert( st, deep_concat( tab[i], sep ) )
				table.insert( st, "}" )
			else
				table.insert( st, tostring( tab[i] ) )
			end
		end
	else
		for key, value in pairs(tab) do
			if type( value ) == 'table' then
				table.insert( st, tostring(key) )
				table.insert( st, "=" )
				table.insert( st, "{" )
				table.insert( st, deep_concat( value, sep ) )
				table.insert( st, "}" )
			else
				table.insert( st, tostring(key) )
				table.insert( st, "=" )
				table.insert( st, tostring( value ) )
			end
		end
	end
	return table.concat(st, sep)
end

function dist( x1, y1, x2, y2 )
	local x = x2-x1
	local y = y2-y1
	return math.sqrt( x*x + y*y )
end

deletion = {}
dead_values = {}

function is_dead( tab, key )
	if not dead_values[tab] then return false end
	return dead_values[tab][key]
end

function delete( tab, key, animated_death )
	if not deletion[tab] then
		deletion[tab] = {}
	end
	if not dead_values[tab] then
		dead_values[tab] = {}
	end
	if not dead_values[tab][key] then
		table.insert( deletion[tab], key )
		if tab[key].sprite then
			if animated_death then
				SetObjAnim( tab[key].sprite, "die", false )
			else
				SetObjDead( tab[key].sprite )
			end
		end
		tab[key].dead = true
		dead_values[tab][key] = true
	end
end

function intersect( obj1, obj2 )
	local o1w, o1h, o2w, o2h
	if type( obj1.size ) == 'table' then
		o1w, o1h = obj1.size[1], obj1.size[2]
	else
		o1w, o1h = obj1.size, obj1.size
	end
	if type( obj2.size ) == 'table' then
		o2w, o2h = obj2.size[1], obj2.size[2]
	else
		o2w, o2h = obj2.size, obj2.size
	end
	return math.abs( obj1.x - obj2.x ) < o1w + o2w and math.abs( obj1.y - obj2.y ) < o1h + o2h
end

function CreateExplosion( x, y )
	CreateFlash( x, y, {1,0.8,0,-0.5}, 128, 0.1 )
	Resume( NewThread( function()
		CreateParticleSystem( "pexplosion", x, y )
		local rings = {}
		local spr
		local angle
		local alpha
		for i=1, math.random(2,3) do
			spr = CreateSprite( "explosion", x, y )
			angle = math.random( 0, 360 ) * math.pi / 180
			alpha = math.random( 0, 5 ) / 10
			SetObjPos( spr, x, y )
			SetObjSpriteSize( spr, alpha )
			SetObjSpriteAngle( spr, angle )
			table.insert( rings, { spr, x, y, alpha } )
		end
		local center = { CreateSprite( "explosion_part", x, y ), x, y, 1 }
		SetObjPos( center[1], x, y )
		while in_game and (#rings > 0 or center )do
			if #rings > 0 then
				for key, value in pairs(rings) do
					value[4] = value[4] + 0.05
					value[3] = value[3] + 0.1
					SetObjPos( value[1], value[2], value[3] )
					SetObjSpriteSize( value[1], 1+value[4] )
					SetObjSpriteColor( value[1], {1,0.8,0,1-value[4]} )
					if value[3] > 400 or value[4] >= 1 then
						SetObjDead( value[1] )
						table.remove( rings, key )
					end
				end
			end
			if center then
				center[2] = center[2] + 0.1
				center[4] = center[4] - 0.05
				SetObjPos( center[1], center[2], center[3] )
				SetObjSpriteSize( center[1], 4-3*center[4] )
				SetObjSpriteColor( center[1], {1,0.8,0,center[4]} )
				if center[3] > 400 or center[4] <= 0 then
						SetObjDead( center[1] )
						center = nil
				end
			end
			Wait(1)
		end
	end ))
end

function CreateBlueExplosion( x, y )
	Resume( NewThread( function()
		CreateParticleSystem( "pexplosionblue", x, y )
		local rings = {}
		local spr
		local angle
		local alpha
		for i=1, math.random(2,3) do
			spr = CreateSprite( "explosion", x, y )
			angle = math.random( 0, 360 ) * math.pi / 180
			alpha = math.random( 0, 5 ) / 10
			SetObjPos( spr, x, y )
			SetObjSpriteSize( spr, alpha )
			SetObjSpriteAngle( spr, angle )
			table.insert( rings, { spr, x, y, alpha } )
		end
		local center = { CreateSprite( "explosion_part_blue", x, y ), x, y, 1 }
		SetObjPos( center[1], x, y )
		while in_game and (#rings > 0 or center) do
			if #rings > 0 then
				for key, value in pairs(rings) do
					value[4] = value[4] + 0.05
					value[3] = value[3] + 0.1
					SetObjPos( value[1], value[2], value[3] )
					SetObjSpriteSize( value[1], 1+value[4] )
					SetObjSpriteColor( value[1], {0.7,0.7,1,1-value[4]} )
					if value[3] > 400 or value[4] >= 1 then
						SetObjDead( value[1] )
						table.remove( rings, key )
					end
				end
			end
			if center then
				center[2] = center[2] + 0.1
				center[4] = center[4] - 0.05
				SetObjPos( center[1], center[2], center[3] )
				SetObjSpriteSize( center[1], 4-3*center[4] )
				SetObjSpriteColor( center[1], {1,1,1,center[4]} )
				if center[3] > 400 or center[4] <= 0 then
						SetObjDead( center[1] )
						center = nil
				end
			end
			Wait(1)
		end
	end ))
end

function GetPassword( pass )
	local pass = string.upper(pass or "nothing")
	if pass == "" then
		pass = "nothing"
	end
	local hsh = hash( pass )
	if hsh == 2623940121 then
		return "level2"
	elseif hsh == 4279505604 then
		return "level3"
	end
	return "level1"
end

function ShutDownEverything( keep_players )
	StopThreads()
	player = nil
	enemies = nil
	bullets = nil
	items = nil
	lightsources = nil
	if not keep_players then
		ResetPlayers()
	end
	ResetEnemies()
	ResetBullets()
	ResetItems()
	ResetLights()
	ResetTimedEvents()
	ResetGUI()
	ProcessBackground = nil
end

--[[
letters = {[0] = "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "Q", "W", "E", "R", "T", "Y", "U", "I", "O", "P", "A", "S", "D" ,"F" ,"G" ,"H" ,"J", "K", "L", "Z", "X", "C", "V", "B", "N", "M"}
--]]

function FinishDeletion()
	for key, value in pairs( deletion ) do
		table.sort( value, function(a,b) return a>b end )
		for i=1,#value do
			--Log( tostring(deep_concat(key, " ")), " ", value[i] )
			table.remove( key, value[i] )
			--Log( tostring(deep_concat(key, " ")) )
		end
		deletion[key] = {}
	end
	dead_values = {}
end

local max = 2^32 -1

local hash_data = {
    0,79764919,159529838,222504665,319059676,
    398814059,445009330,507990021,638119352,
    583659535,797628118,726387553,890018660,
    835552979,1015980042,944750013,1276238704,
    1221641927,1167319070,1095957929,1595256236,
    1540665371,1452775106,1381403509,1780037320,
    1859660671,1671105958,1733955601,2031960084,
    2111593891,1889500026,1952343757,2552477408,
    2632100695,2443283854,2506133561,2334638140,
    2414271883,2191915858,2254759653,3190512472,
    3135915759,3081330742,3009969537,2905550212,
    2850959411,2762807018,2691435357,3560074640,
    3505614887,3719321342,3648080713,3342211916,
    3287746299,3467911202,3396681109,4063920168,
    4143685023,4223187782,4286162673,3779000052,
    3858754371,3904687514,3967668269,881225847,
    809987520,1023691545,969234094,662832811,
    591600412,771767749,717299826,311336399,
    374308984,453813921,533576470,25881363,
    88864420,134795389,214552010,2023205639,
    2086057648,1897238633,1976864222,1804852699,
    1867694188,1645340341,1724971778,1587496639,
    1516133128,1461550545,1406951526,1302016099,
    1230646740,1142491917,1087903418,2896545431,
    2825181984,2770861561,2716262478,3215044683,
    3143675388,3055782693,3001194130,2326604591,
    2389456536,2200899649,2280525302,2578013683,
    2640855108,2418763421,2498394922,3769900519,
    3832873040,3912640137,3992402750,4088425275,
    4151408268,4197601365,4277358050,3334271071,
    3263032808,3476998961,3422541446,3585640067,
    3514407732,3694837229,3640369242,1762451694,
    1842216281,1619975040,1682949687,2047383090,
    2127137669,1938468188,2001449195,1325665622,
    1271206113,1183200824,1111960463,1543535498,
    1489069629,1434599652,1363369299,622672798,
    568075817,748617968,677256519,907627842,
    853037301,1067152940,995781531,51762726,
    131386257,177728840,240578815,269590778,
    349224269,429104020,491947555,4046411278,
    4126034873,4172115296,4234965207,3794477266,
    3874110821,3953728444,4016571915,3609705398,
    3555108353,3735388376,3664026991,3290680682,
    3236090077,3449943556,3378572211,3174993278,
    3120533705,3032266256,2961025959,2923101090,
    2868635157,2813903052,2742672763,2604032198,
    2683796849,2461293480,2524268063,2284983834,
    2364738477,2175806836,2238787779,1569362073,
    1498123566,1409854455,1355396672,1317987909,
    1246755826,1192025387,1137557660,2072149281,
    2135122070,1912620623,1992383480,1753615357,
    1816598090,1627664531,1707420964,295390185,
    358241886,404320391,483945776,43990325,
    106832002,186451547,266083308,932423249,
    861060070,1041341759,986742920,613929101,
    542559546,756411363,701822548,3316196985,
    3244833742,3425377559,3370778784,3601682597,
    3530312978,3744426955,3689838204,3819031489,
    3881883254,3928223919,4007849240,4037393693,
    4100235434,4180117107,4259748804,2310601993,
    2373574846,2151335527,2231098320,2596047829,
    2659030626,2470359227,2550115596,2947551409,
    2876312838,2788305887,2733848168,3165939309,
    3094707162,3040238851,2985771188,
}

local function xor(a, b)
    local calc = 0    

    for i = 32, 0, -1 do
	local val = 2 ^ i
	local aa = false
	local bb = false

	if a == 0 then
	    calc = calc + b
	    break
	end

	if b == 0 then
	    calc = calc + a
	    break
	end

	if a >= val then
	    aa = true
	    a = a - val
	end

	if b >= val then
	    bb = true
	    b = b - val
	end

	if not (aa and bb) and (aa or bb) then
	    calc = calc + val
	end
    end

    return calc
end

local function lshift(num, left)
    local res = num * (2 ^ left)
    return res % (2 ^ 32)
end

local function rshift(num, right)
    local res = num / (2 ^ right)
    return math.floor(res)
end

function hash(str)
    local count = string.len(tostring(str))
    local crc = max
    
    local i = 1
    while count > 0 do
	local byte = string.byte(str, i)

	crc = xor(lshift(crc, 8), hash_data[xor(rshift(crc, 24), byte) + 1])

	i = i + 1
	count = count - 1
    end

    return crc
end
