#ifndef __OBJECT_DYNAMIC_H_
#define __OBJECT_DYNAMIC_H_

#include "object_physic.h"
#include "object_environment.h"

enum ObjectMovementType { omtIdling, omtWalking, omtSitting, omtJumping, omtLanding/*, omtAimingUp, omtAimingDown, omtWalkingAimingUp*/, omtDropSitting, omtMovingToWaypoint, omtDodging, omtLying, omtRolling };
enum ObjectFacingType { ofNormal, ofMoonwalking, ofFixed };


extern ObjEnvironment* default_environment;

class ObjDynamic : public ObjPhysic
{
public:
	CAABB old_aabb;
	Vector2 vel;
	Vector2 acc;
	Vector2 gravity;
	Vector2 own_vel; //Only once each update, does not change speed
	TrajectoryType trajectory;
	float t_value;
	float t_param1;
	float t_param2;
	float max_x_vel;
	float max_y_vel;
	ObjPhysic* suspected_plane;
	ObjPhysic* drop_from;			//   ,  ,    .
	ObjEnvironment* env;
	//ObjEnvironment* old_env;
	float shadow_width;
	bool drops_shadow;
	bool facing_left;
	float lifetime;
	float mass;

	float walk_acc;
	float jump_vel;

	bool ghostlike;			//  "",    
							//      solid-

	bool movementDirectionX;		//   


	ObjectMovementType movement;		//   
	ObjectFacingType facing;

	BYTE physPushes;
	


	// 
	// on_plane		16		   
	// bullet		32		 - 
	// dropping		128		   
	__INLINE BYTE IsOnPlane()		{ return physFlags & 16; }
	__INLINE BYTE IsBullet()		{ return physFlags & 32; }
	__INLINE BYTE IsDropping()		{ return physFlags & 128; }

	__INLINE void SetOnPlane()		{ physFlags |= 16; }
	__INLINE void SetBullet()		{ physFlags |= 32; }
	__INLINE void SetDropping()		{ physFlags |= 128; drop_from = suspected_plane; }

	__INLINE void ClearOnPlane()	{ physFlags &= ~16; }
	__INLINE void ClearBullet()		{ physFlags &= ~32; }
	__INLINE void ClearDropping()	{ physFlags &= ~128; }

	__INLINE BYTE IsPushedLeft()		{ return physPushes & 1; }
	__INLINE BYTE IsPushedRight()		{ return physPushes & 2; }
	__INLINE BYTE IsPushedUp()		{ return physPushes & 4; }
	__INLINE BYTE IsPushedDown()		{ return physPushes & 8; }

	__INLINE void SetPushedLeft()		{ physPushes |= 1; }
	__INLINE void SetPushedRight()		{ physPushes |= 2; }
	__INLINE void SetPushedUp()		{ physPushes |= 4; }
	__INLINE void SetPushedDown()		{ physPushes |= 8; }

	__INLINE void ClearPushes()	{ physPushes = 0; }

	ObjDynamic()
	{
		this->SetDynamic();

		suspected_plane = NULL;
		drop_from = NULL;
		env = default_environment;
		//old_env = default_environment;

		trajectory = pttLine;
		t_param1 = 0.0f;
		t_param2 = 0.0f;

		mass = 1.0f;

		max_x_vel = 0.0f;
		max_y_vel = 0.0f;

		walk_acc = 0.0f;
		jump_vel = 0.0f;
		drops_shadow = false;
		shadow_width = 0.25f;

		lifetime = 0.0f;
		facing_left = false;

		movementDirectionX = true;

		movement = omtIdling;
		facing = ofNormal;

		physPushes = 0;
		

		ghostlike = false;
	}

	void SetFacing( bool mirrored )
	{
		if ( this->childrenConnection && facing_left ^ mirrored )
			this->childrenConnection->Event( ObjectEventInfo( eventFacingChange ) );
		facing_left = mirrored;
		if ( !sprite ) return;
		if ( facing == ofFixed ) return;
		if ( mirrored )
			ofMoonwalking == facing ? this->sprite->ClearMirrored() : this->sprite->SetMirrored();
		else
			ofMoonwalking == facing ? this->sprite->SetMirrored() : this->sprite->ClearMirrored();
	}

	bool GetFacing()
	{
		return facing_left;
	}

	void PhysProcess();

	bool CheckSuspectedPlane();
	void CheckUnderlyngPlane();
	void DropShadow();

	void SetEnv(ObjEnvironment* e);
};


#endif // __OBJECT_DYNAMIC_H_
