#include "StdAfx.h"
#include "object_particle_system.h"
#include "../../misc.h"
#include "../phys/phys_misc.h"
#include "object.h"
#include "../../resource_mgr.h"

//////////////////////////////////////////////////////////////////////////

extern ResourceMgr<Proto> * protoMgr;
extern ResourceMgr<Texture> * textureMgr;

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

GameObject* CreateParticleSystem( const char * proto_name, Vector2 coord, UINT emitter, int mode )
{	
	const Proto* proto = protoMgr->GetByName( proto_name, "particles/" );
	if (proto == NULL)
	{
		return NULL;
	}

	GameObject* obj = new GameObject();
	
	if (LoadObjectFromProto(proto, obj))
	{
		DELETESINGLE(obj);
		return NULL;
	}
	obj->type = objParticleSystem;
	
	obj->ParticleSystem = new CParticleSystem();

	obj->ParticleSystem->LoadFromProto(proto);
	obj->ParticleSystem->Info.p = coord;
	obj->ParticleSystem->system_z = proto->z;

	obj->ParticleSystem->emitter = emitter;
	obj->ParticleSystem->mode = mode;
	if ( emitter && (mode & 4) )
	{
		GameObject* emit_obj = GetGameObject( emitter );
		if ( emit_obj )
		{
			CAABB* emit_aabb = &emit_obj->aabb;
			Vector2 * pts = new Vector2[2];
			pts[1] = Vector2(emit_aabb->p.x-emit_aabb->W/*/2*/,emit_aabb->p.y);
			pts[0] = Vector2(emit_aabb->p.x+emit_aabb->W/*/2*/,emit_aabb->p.y);
			obj->ParticleSystem->SetGeometry(pts, 2);
		}
	}
	
	obj->ParticleSystem->Init();
	
	AddObject(obj);
	return obj;

}
