#include "StdAfx.h"
#include "particle_system.h"
#include "../misc.h"
#include "phys/phys_misc.h"
#include "objects/object.h"
#include "objects/object_physic.h"
#include "objects/object_player.h"
#include "objects/object_dynamic.h"
#include "player.h"
#include "camera.h"

extern GameObject* attached_object;

/*
Ðåæèìû:
1 - ñîçäàíèå ïî âñåìó aabb îáúåêòà.
2 - ïðèìåíÿòü ãðàâèòàöèþ.
3 - "ñíåã", ñîçäàíèå ïî îòðåçêó øèðèíû ñòàðòîâîé ïîçèöèè îáúåêòà âíèç.
*/

Vector2 wind;

void SetParticlesWind( Vector2 new_wind )
{
	wind = new_wind;
}

Vector2 GetParticlesWind()
{
	return wind;
}

template<typename T>
__INLINE void _swap(int i, int j, T* arr)
{
	T t = arr[i];
	arr[i] = arr[j];
	arr[j] = t;
}

__INLINE void CParticleSystem::swapParticles(int i, int j)
{
	_swap(i, j, particles);
	_swap(i, j, pp);
	_swap(i, j, pcolor);
	_swap(i, j, pz);
}


void CParticleSystem::Init()
{
	this->particles = new CParticle [Info.MaxParticles];
	this->pp = new Vector2[Info.MaxParticles];
	this->pcolor = new RGBAf[Info.MaxParticles];
	this->pz = new float[Info.MaxParticles];
}

bool AreaMovement( Vector2& coord, Vector2 vel, vector<CAABB>* area )
{
#ifndef _DEBUG
	if ( area == NULL || area->size() == 0 )
#endif
	{
		coord += vel;
		return true;
	}
	Vector2 pnt = coord + vel;
	for ( vector<CAABB>::iterator it = area->begin(); it != area->end(); it++ )
	{
		if ( it->PointInCAABB(pnt.x, pnt.y) )
		{
			coord += vel;
			return true;
		}
	}
	return false;
}

bool CParticleSystem::Update()
{
#ifdef _DEBUG_DISABLE_PARTICLES_UPDATE
	return false;
#endif
	// Here integrating and updating values of active particles
	for(size_t i=0; i < Info.ParticlesActive; i++)
	{
		CParticle& pi = particles[i];
		
		pi.age++;
		if (pi.age >= pi.life)
		{
			swapParticles(i, Info.ParticlesActive-1);
			Info.ParticlesActive--;
			i--;
			continue;
		}
		Vector2& p = pp[i];
		RGBAf& color = pcolor[i];
		
		pi.v += Info.gravity;
		pi.old_p = p;

#ifndef _DEBUG
		if ( drop_area != NULL && !pi.is_on_plane && drop_area->size() > 0 )
#else
		if ( false )
#endif
		{
			for ( vector<CAABB>::iterator it = drop_area->begin(); it != drop_area->end(); it++ )
			{
				if ( it->PointInCAABB(p.x, p.y) )
				{
					if ( rand() % 100 > 90 )
					{
						pi.is_on_plane = true;
						pz[i] = SHADOW_Z;
					}
					break;
				}
			}
		}

		if ( pi.is_on_plane )
		{
			color += (pi.dc*0.2f);
			pi.size += pi.dsize;
			continue;
		}

		switch ( Info.trajectory )
		{
			case pttRipple:
				{
					pi.parameter += 0.2f;

					// TODO: Íå çíàþ, ÷òî â äàííîì ñëó÷àå ëó÷øå: sin èëè FastSin? Ïðîôàéëåð ïîêàçûâàåò, ÷òî 
					// sin ÷óòî÷êó ëó÷øå. Âîçìîæíî, íà äðóãèõ ïðîöåññîðàõ ðåçóëüòàò áóäåò äðóãèì.
					// Íó õîòü InvLengthQ çàìåòíî óñêîðÿåò.
					//float tr = Info.t_param1 * FastSin<false>(pi.parameter*Info.t_param2) * pi.v.InvLengthQ();
					float tr = Info.t_param1 * FastSin<false>(pi.parameter*Info.t_param2) * InvSqrt( pi.v.x*pi.v.x + pi.v.y*pi.v.y );
					//float tr = Info.t_param1 * sin(pi.parameter*Info.t_param2) * pi.v.InvLengthQ();
					
					float tr_x = -pi.v.x * tr;
					float tr_y = pi.v.y * tr;
					if (!AreaMovement( p, (pi.v + Vector2( tr_x, tr_y ))*0.2f, area ))
						AreaMovement( p, Info.gravity, area );
					color += (pi.dc*0.2f);
					pi.size += pi.dsize;
					break;
				}
			case pttTwist:
				{
					Vector2 center_point = p;
					Vector2 old_center_point = center_point;
					if ( !dead ) 
						center_point = old_center_point = Info.p;
					if ( mode & 2 )
					{
						GameObject* go = GetGameObject( emitter );
						if ( go )
						{
							center_point = go->aabb.p;
							if ( go->IsPhysic() && ((ObjPhysic*)go)->IsDynamic() )
							old_center_point = center_point - ((ObjDynamic*)go)->vel;
						}
					}
					pi.parameter += 0.2f*Info.t_param1;
					
					//float ps = sin(pi.parameter);
					float ps = FastSin<false>(pi.parameter);
					// TODO: Ñþäà FastCos, êîòîðîãî íåò.
					float pc = cos(pi.parameter);

					if (!AreaMovement( p, center_point + (0.2f*pi.v.Length() + (p - center_point).Length()) * Vector2(ps, pc) - p, area))
						AreaMovement( p, Info.gravity, area );
					color += (pi.dc*0.2f);
					pi.size += pi.dsize;
					break;
				}
			case pttSine:
				{
					pi.parameter += 0.2f;
					Vector2 perp = Vector2();
					if ( pi.v.Length() > 0 ) perp = pi.v.GetPerpendicular().Normalized();
					Vector2 tr = perp * Info.t_param1 * FastSin<false>(pi.parameter*Info.t_param2);
					//Vector2 tr(pi.v.GetPerpendicular());
					//tr.Normalize();
					//tr *= Info.t_param1 * FastSin<false>(pi.parameter*Info.t_param2);

					if (!AreaMovement( p, (pi.v + tr)*0.2f, area ))
						AreaMovement( p, Info.gravity, area );
					color += (pi.dc*0.2f);
					pi.size += pi.dsize;
					break;
				}
			case pttCosine:
				{
					pi.parameter += 0.2f;
					Vector2 perp = Vector2();
					if ( pi.v.Length() > 0 ) perp = pi.v.GetPerpendicular().Normalized();
					Vector2 tr = perp * Info.t_param1 * FastSin<false>(pi.parameter*Info.t_param2);
					//Vector2 tr(pi.v.GetPerpendicular());
					//tr.Normalize();
					//tr *= Info.t_param1 * FastSin<false>(pi.parameter*Info.t_param2);

					if (!AreaMovement( p, (pi.v + tr)*0.2f, area ))
						AreaMovement( p, Info.gravity, area );
					color += (pi.dc*0.2f);
					pi.size += pi.dsize;
					break;
				}
			case pttPseudoDepth:
				{
					pi.size += pi.parameter*0.2f;
					if (pi.size > Info.t_param2 || pi.size < -Info.t_param2) pi.parameter = 0;
					if (!AreaMovement( p, pi.v*0.2f, area))
						AreaMovement( p, Info.gravity, area );
					color += (pi.dc*0.2f);
					break;
				}
			case pttRandom:
				{
					//Vector2 tr = pi.v.GetPerpendicular().Normalized() * Random_Float(-pi.parameter, pi.parameter);
					Vector2 perp = Vector2();
					Vector2 norm = Vector2();
					if ( pi.v.Length() > 0 ) 
					{
						perp = pi.v.GetPerpendicular().Normalized();
						norm = pi.v.Normalized();
					}
					Vector2 tr = perp * Random_Float(-Info.t_param1, Info.t_param1);
					
					Vector2 tr2 = norm * Random_Float(-Info.t_param2, Info.t_param2);;
					pi.v += tr;
				}
			default:
				{
					if (!AreaMovement( p,  pi.v*0.2f, area ))
						AreaMovement( p, Info.gravity, area );
					color += (pi.dc*0.2f);
					pi.size += pi.dsize;
					break;
				}
		}
		if (Info.wind_affected)
		{
			AreaMovement( p, wind*0.2f, area );
		}
	}


	float np = Info.emission * 0.2f;	// + uncreated;

	if (Info.life != -1)
	{
		Info.age++;
		if (Info.age >= Info.life)
		{
			this->dead = true;
			return true;
		}
	}

	size_t t = Info.ParticlesActive;
	for (size_t i = t; i < t + np; i++)
	{
		CreateParticle();
	}
	return true;
}

void CParticleSystem::Render()
{
#ifdef _DEBUG_DISABLE_PARTICLES_DRAW
	return;
#endif
	if (Info.ParticlesActive == 0)
		return;

	for(size_t i=0;i<Info.ParticlesActive;i++)
	{
		Vector2 norm = particles[i].old_p - pp[i];
		if ( norm.Length() > 0 )
			norm = norm.Normalized();
		else
			norm = Vector2();
		Vector2 trace = pp[i] + ( norm * particles[i].trace);
		float halfsize = particles[i].size * 0.5f;
		if ( trace != pp[i] )
		{
			RenderLine( pp[i].x + halfsize, pp[i].y + halfsize,
						trace.x + halfsize, trace.y + halfsize,
						pz[i]-0.0001f, pcolor[i] );
		}
	}
	if (!texture)
	{
		glPushAttrib(GL_TEXTURE_BIT | GL_POINTS);
		glEnable(GL_POINTS);
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_TEXTURE);
		glPointSize(particles[0].size);
		glBegin(GL_POINTS);
		for(size_t i=0;i<Info.ParticlesActive;i++)
		{
			//pcolor[i].glSet();
			glColor4fv(&pcolor[i].r);
			glVertex3f(pp[i].x, pp[i].y, pz[i]);
		}
		glEnd();
		glPopAttrib();
	}
	else
	{
//#ifdef RENDER_PARTICLES_WITH_STD_RENDER
#if 1
		
		FrameInfo* fi = this->texture->frame;
		ASSERT(fi);
		size_t frameNum = 0;

		coord2f_t* fsizes = new coord2f_t[Info.ParticlesActive];
		coord2f_t** tex_coords = new coord2f_t*[Info.ParticlesActive];

		for(size_t i=0;i<Info.ParticlesActive;i++)
		{
			const CParticle& par = particles[i];
			fsizes[i].x = fsizes[i].y = par.size;

			if (this->texture->framesCount > 1)
			{
				// Âûáîð êàäðà àíèìàöèè
				frameNum = (size_t)((par.age / par.life) * this->texture->framesCount);
				if (frameNum >= this->texture->framesCount) frameNum = this->texture->framesCount-1;
			}
			
			tex_coords[i] = fi[frameNum].coord;
		}
		RenderSprite(pp, pz, fsizes, tex_coords, texture, pcolor, Info.ParticlesActive);

		DELETEARRAY(fsizes);
		DELETEARRAY(tex_coords);
#else

		glPushAttrib(GL_TEXTURE_BIT | GL_POINTS);
		if ( Info.fancy_blending )
			glBlendFunc(GL_SRC_ALPHA, GL_ONE);

		glDisable(GL_POINTS);
		glEnable(GL_TEXTURE_2D);
		
		glTranslatef(0.0f, 0.0f, 0.0f);
		glBindTexture(GL_TEXTURE_2D, texture->tex);

		glBegin(GL_QUADS);


		FrameInfo* fi = this->texture->frame;
		ASSERT(fi);
		size_t frameNum = 0;
				
		for(size_t i=0;i<Info.ParticlesActive;i++)
		{
			const CParticle& pp = particles[i];
			//particles[i].c.glSet();
			glColor4fv(&pcolor[i].r);
			
			if (this->texture->framesCount > 1)
			{
				// Âûáîð êàäðà àíèìàöèè
				frameNum = (size_t)((pp.age / pp.life) * this->texture->framesCount);
				if (frameNum >= this->texture->framesCount) frameNum = this->texture->framesCount-1;
			}

			glTexCoord2fv((const GLfloat*)(fi[frameNum].coord + 0)); glVertex3f(floor(pp.p.x - 0.5f*pp.size + 0.5f),	floor(pp.p.y - 0.5f*pp.size + 0.5f), this->z);
			glTexCoord2fv((const GLfloat*)(fi[frameNum].coord + 1)); glVertex3f(floor(pp.p.x + 0.5f*pp.size + 0.5f),	floor(pp.p.y - 0.5f*pp.size + 0.5f), this->z);
			glTexCoord2fv((const GLfloat*)(fi[frameNum].coord + 2)); glVertex3f(floor(pp.p.x + 0.5f*pp.size + 0.5f),	floor(pp.p.y + 0.5f*pp.size + 0.5f), this->z);
			glTexCoord2fv((const GLfloat*)(fi[frameNum].coord + 3)); glVertex3f(floor(pp.p.x - 0.5f*pp.size + 0.5f),	floor(pp.p.y + 0.5f*pp.size + 0.5f), this->z);
		}
		glEnd();
		glPopAttrib();

		if ( Info.fancy_blending )
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
#endif	// RENDER_PARTICLES_WITH_STD_RENDER
	}


	return;
}

void CParticleSystem::SetGeometry( Vector2 * points, int numPoints )
{
	Info.geom = points;
	Info.GeomNumPoints = numPoints;
}

extern float CAMERA_TOP;
extern float CAMERA_BOTTOM;
extern float CAMERA_LEFT;
extern float CAMERA_RIGHT;

CParticle	*CParticleSystem::CreateParticle()
{
	if ( dead ) return NULL;

	size_t t = Info.ParticlesActive;
	size_t i = t;
	CParticle& pi = particles[i];
	Vector2& p = pp[i];
	RGBAf& color = pcolor[i];
	{
		if (i >= Info.MaxParticles)
			return NULL;
		if (Info.ParticlesActive >= Info.MaxParticles)
			return NULL;
		Info.ParticlesActive++;
	
		Vector2 emit_point = Info.p;
		if ( emitter )
		{
			GameObject * emit_obj = GetGameObject( emitter );
			if ( emit_obj )
			{
				if ( mode & 8 )
				{
					emit_point = emit_obj->aabb.p + this->Info.p;
				}
				else
				{
					float dx = (rand()%((int)(emit_obj->aabb.W*100)))/100.0f;
					float dy = (rand()%((int)(emit_obj->aabb.H*100)))/100.0f;
					emit_point = Vector2(emit_obj->aabb.p.x-(emit_obj->aabb.W/2)+dx,
								         emit_obj->aabb.p.y-(emit_obj->aabb.H/2)+dy);
				}
			}
			else
			{
				this->dead = true;
				return NULL;
			}
		}
		if ( mode & 16 )
		{
			ObjDynamic* op = (ObjDynamic*)attached_object;
			float vx = 0;
			float vy = 0;
			if ( op )
			{
				vx += op->vel.x;
				vy += op->vel.y;
			}
			emit_point = Vector2( Random_Float( CAMERA_LEFT+vx-320, CAMERA_RIGHT+vx+320 ), CAMERA_TOP-Info.startsize+Info.sizevar+vy-240 );
			if ( emit_point.x > CAMERA_RIGHT || emit_point.x < CAMERA_LEFT )
				emit_point.y += Random_Float( -100.0f, (CAMERA_BOTTOM - CAMERA_TOP)+100.0f );
			else
				emit_point.y -= Random_Float( 0.0f, 100.0f );
		}
		pi.life  = Random_Int(Info.plife ,Info.plife + Info.plifevar);
		if (pi.life < 0 )
			pi.life = Info.plife;
		pi.age = 0;
		pi.parameter = Random_Float( Info.min_param, Info.max_param );

		if ( area != NULL && area->size() > 0 )
		{
			bool in_area = false;
			for ( vector<CAABB>::iterator it = area->begin(); it != area->end(); it++ )
			{
				if ( it->PointInCAABB( emit_point.x, emit_point.y ) )
				{
					in_area = true;
					break;
				}
			}
			if (!in_area)
				return NULL;
		}

		if (Info.geom != NULL && Info.GeomNumPoints > 1)
		{
			int sr = Random_Int(0, Info.GeomNumPoints-2);
			p = Info.geom[sr] + (Info.geom[sr+1]-Info.geom[sr])*Random_Float(0.0f, 1.0f);
			float angle = PI * Random_Float(Info.min_angle, Info.max_angle) / 180;
			float speed = Random_Float(Info.max_speed, Info.max_speed);
			Vector2 norm = (Info.geom[sr]-Info.geom[sr+1]).GetPerpendicular();
			if ( norm.Length() > 0 )
				norm = norm.Normalized();
			else
				norm = Vector2();
			pi.v = norm * speed;
			pi.v.x *= cos(angle);
			pi.v.y *= sin(angle);
		}
		else
		{
			p = emit_point;
			float angle = PI * Random_Float(Info.min_angle, Info.max_angle) / 180;
			float speed = Random_Float(Info.max_speed, Info.max_speed);
			if ( Info.trajectory == pttPseudoDepth )
			{
				float angle2 = PI * Random_Float( -Info.t_param1, Info.t_param1 ) / 180;
				pi.parameter = speed * sin(angle2);
				pi.v = Vector2( speed * cos(angle) * cos(angle2), speed * sin(angle) * cos(angle2) );				
			}
			else
			{
				pi.v = Vector2( speed * cos(angle), speed*sin(angle) );
			}
		}
		pi.old_p = p;

		pi.size = Random_Float(Info.startsize, Info.startsize + Info.sizevar);
		pi.dsize = 5*(Info.endsize - Info.startsize) / pi.life;

		color.a = Random_Float(Info.sc.a ,Info.sc.a + Info.vc.a);
		color.r = Random_Float(Info.sc.r ,Info.sc.r + Info.vc.r);
		color.g = Random_Float(Info.sc.g ,Info.sc.g + Info.vc.g);
		color.b = Random_Float(Info.sc.b ,Info.sc.b + Info.vc.b);

		pi.dc = (Info.ec - color)/(pi.life/5.0f);
		pi.trace = Random_Float(Info.min_trace, Info.max_trace);
		pi.is_on_plane = false;

		pz[i] = system_z;
}
	return &pi;
}



CParticleSystem::~CParticleSystem()
{
	if ( area )	DELETESINGLE(area);
	if ( drop_area ) DELETESINGLE(drop_area);
	DELETEARRAY(particles);
	DELETEARRAY(this->Info.geom);

	DELETEARRAY(pp);
	DELETEARRAY(pz);
	DELETEARRAY(pcolor);
}

void CParticleSystem::LoadFromProto( const Proto* proto )
{
	extern ResourceMgr<Texture> * textureMgr;

	Info.sc = proto->start_color;
	Info.ec = proto->end_color;
	Info.vc = proto->var_color;

	texture = textureMgr->GetByName(proto->texture);
	
	Info.emission = proto->emission;
	Info.MaxParticles = (size_t)proto->max_particles;
	Info.lifemin = (int)proto->particle_life_min;
	Info.lifemax = (int)proto->particle_life_max;

	Info.min_speed = proto->min_speed;
	Info.max_speed = proto->max_speed;
	Info.min_angle = proto->min_angle;
	Info.max_angle = proto->max_angle;
	Info.min_param = proto->min_param;
	Info.max_param = proto->max_param;
	Info.min_trace = proto->min_trace;
	Info.max_trace = proto->max_trace;
	Info.trajectory = (TrajectoryType)proto->trajectory;
	Info.t_param1 = proto->trajectory_param1;
	Info.t_param2 = proto->trajectory_param2;
	Info.wind_affected = proto->affected_by_wind == 1;
	Info.gravity = Vector2( proto->gravity_x, proto->gravity_y );

	Info.startsize = proto->start_size;
	Info.sizevar = proto->size_variability;
	Info.endsize = proto->end_size;

	Info.plife = (int)proto->particle_life * 5;
	Info.plifevar = (int)proto->particle_life_var * 5;

	Info.life = (int)proto->system_life;
	Info.life *= Info.life == -1 ? 1 : 5;
}
