#include "StdAfx.h"

#include "../misc.h"
#include "../cfile.h"
#include "font.h"

#include "../resource_mgr.h"

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

extern char path_fonts[MAX_PATH];
extern ResourceMgr<Texture> * textureMgr;

//////////////////////////////////////////////////////////////////////////

map<string, IFont*> fonts;	//  

//////////////////////////////////////////////////////////////////////////

UCHAR *IFont::buffer = NULL;
size_t IFont::buf_len = 0;

//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////



//    
// in_name -    
// out_name -        
bool LoadTextureFont(const char* in_name, const char* out_name)
{
	if(!in_name || !out_name)
	{
		sLog(DEFAULT_LOG_NAME, LOG_WARNING_EV, "Incorrect in_name or out_name for loading windows font");
		return false;
	}

	if (FontByName(out_name))
	{
		sLog(DEFAULT_LOG_NAME, LOG_WARNING_EV, " .");
		return false;
	}

	char* font_file = new char[strlen(path_fonts) + strlen(in_name) + 1];
	memset(font_file, '\0', strlen(path_fonts) + strlen(in_name) + 1);
	sprintf(font_file, "%s%s", path_fonts, in_name);

	TexFont* font = new TexFont;
	font->filename = font_file;

	DELETEARRAY(font_file);

	if (!font->Load())
	{
		DELETESINGLE(font);
		return false;
	}

	fonts[string(out_name)] = font;

	return true;
}

//   .
void RecoverFonts()
{
	map<string, IFont*>::iterator it;
	for(it = fonts.begin();
		it != fonts.end();
		it++)
	{
		IFont* font = it->second;
		font->Recover();
	}
}

void FreeFont(const char* font_name)
{
	map<string, IFont*>::iterator it = fonts.find(string(font_name));
	if (it != fonts.end())
	{
		IFont* font = it->second;
		DELETESINGLE(font);
		fonts.erase(it);
	}
}



// free all fonts
void FreeFonts(void)
{
	map<string, IFont*>::iterator it;
	for(it = fonts.begin();
		it != fonts.end();
		it++)
	{
		IFont* font = it->second;
		DELETESINGLE(font);
	}
	fonts.clear();
}



// get font by name
IFont* FontByName(const char* font_name)
{
	map<string, IFont*>::iterator it = fonts.find(string(font_name));
	if (it != fonts.end())
	{
		return it->second;
	}

	return NULL;
}

//////////////////////////////////////////////////////////////////////////


void IFont::PrintMultiline( const char *text, CAABB &area, int cursor )
{
	if (text == NULL)
		return;

	size_t br = 1;				//   .
	size_t back = 0;			
	int x_size = 0;				//   
	int y_size = 0;
	float current_shift = 0;	//    . ,   
	float old_x = this->p.x;
	float old_y = this->p.y;
	size_t max_br;

	size_t left_text_len = 0;	//    

	ControlSeqParser::CSeqType seq_type = ControlSeqParser::ENone;
	size_t seq_shift = 0;

	size_t text_len = strlen( text );
	SetBufLen(text_len + 2);

	while ( br <= text_len )
	{
		for (x_size = 0; x_size < 2*area.W - current_shift && br <= text_len; br++)
		{	
			seq_type = ControlSeqParser::CheckControlSeq(&text[br-1], seq_shift);
			if (seq_type != ControlSeqParser::ENone)
				break;

			x_size += GetWidthAndDist(text[br-1]);
		}

		br += seq_shift;
		left_text_len = br-seq_shift-back-1;

		switch(seq_type)
		{
		case ControlSeqParser::EColor:
			{
				//   
				if (br-back-1 > 0)
				{
					memcpy(buffer, text + back, left_text_len);
					buffer[left_text_len] = 0;
					Draw(buffer, cursor - back);
				}

				this->p.x += x_size;
				current_shift += x_size;

				ControlSeqParser::GetColor(&text[br - seq_shift - 1], this->tClr);

				back = br-1;
				continue;
			}
			break;
		case ControlSeqParser::ENewLine: break;
		default:
			{
				max_br = br;
				while ( text[br-1] != 32 && br-1 > back && br != text_len+1  )	//   
					br--;
				if ( br-1 == back ) br = max_br; // .

				left_text_len = br-seq_shift-back-1;
			}
		}

		seq_type = ControlSeqParser::ENone;

		memcpy(buffer, text + back, left_text_len);
		buffer[left_text_len] = 0;
		Draw(buffer, cursor - back);

		y_size = GetStringHeight((char*)buffer);

		area = CAABB( area.Left(), area.Top(), area.Right(), max( area.Bottom(), area.Top() + this->p.y - old_y + y_size  ) );

		this->p.y += y_size * 1.5f;
		this->p.x = old_x;
		current_shift = 0;

		back = br-1;
		while ( text[back] == 32 && back < text_len )	//     
		{
			back++;
			br++;
		}

	}

	this->p.y = old_y;
	this->p.x = old_x;

}

//-------------------------------------------//
//				TexFont stuff					 //
//-------------------------------------------//
TexFont::TexFont()
{
	FontImageName = NULL;
	dist = TEXFONT_DEFAULT_DISTANCE;
	tClr = RGBAf(1.0f, 1.0f, 1.0f, 1.0f);
	// TODO:   bbox
}

TexFont::~TexFont()
{
	if (FontImageName != NULL)
		delete [] FontImageName;
}

//    
//    filename
bool TexFont::Load()
{
	if (filename == "")
		return false;
	CFile file;

	if (!file.Open(filename.c_str(), CFile::OPEN_MODE_READ))
	{
		sLog(DEFAULT_LOG_NAME, LOG_WARNING_EV, "Can't Load Font %s: file  couldn't be opened.", filename.data());
		return false;
	}

	file.ReadLine(FontImageName);
	tex = textureMgr->GetByName(FontImageName);
	if (!this->tex)
	{
		file.Close();
		return false;
	}
	file.Read(bbox, sizeof(bbox));
	float y;

	// TODO:   - .       
	//   ,     .
	for (size_t i = 0; i < sizeof(bbox) / sizeof(bbox[0]); i++)
	{
		y = (float)tex->height - bbox[i].Max.y;
		bbox[i].Max.y = (float)tex->height - bbox[i].Min.y;
		bbox[i].Min.y = y;

		width[i] = (BYTE)(bbox[i].Max.x - bbox[i].Min.x);
		height[i] = (BYTE)(bbox[i].Max.y - bbox[i].Min.y);
	}
	file.Close();

	return true;
}

void TexFont::Recover()
{
	this->Load();
}

bool TexFont::SaveToFile()
{
	if (filename == "")
		return false;
	CFile			file;
	file.Open(filename, CFile::OPEN_MODE_WRITE);
	file.Write(FontImageName, (size_t)strlen(FontImageName));
	file.WriteByte((byte)0x00);
	file.Write(bbox, sizeof(bbox));
	file.Close();
	return true;
}

void TexFont::Draw(const UCHAR *text)
{
	unsigned char c;
	float x = p.x;
	float y = p.y;
	for (const UCHAR* t = text; *t; t++)
	{
		if (*t < 32)
			continue;
		c = *t - 32;
		if ( outline.a > 0 )
		{
			RenderSprite(x-1, y, z, bbox[c].Min.x, bbox[c].Min.y, bbox[c].Max.x, bbox[c].Max.y, tex, false, outline);
			RenderSprite(x+1, y, z, bbox[c].Min.x, bbox[c].Min.y, bbox[c].Max.x, bbox[c].Max.y, tex, false, outline);
			RenderSprite(x, y-1, z, bbox[c].Min.x, bbox[c].Min.y, bbox[c].Max.x, bbox[c].Max.y, tex, false, outline);
			RenderSprite(x, y+1, z, bbox[c].Min.x, bbox[c].Min.y, bbox[c].Max.x, bbox[c].Max.y, tex, false, outline);
		}
		RenderSprite(x, y, z, bbox[c].Min.x, bbox[c].Min.y, bbox[c].Max.x, bbox[c].Max.y, tex, false, tClr);
		x +=width[c] + dist;
	}
}

extern UINT current_time;

void TexFont::Draw(const UCHAR *text, int cursor)
{
	unsigned char c;
	float x = p.x;
	float y = p.y;
	int pos = 0;
	for (const UCHAR* t = text; *t; t++)
	{
		if (*t < 32)
			continue;
		c = *t - 32;
		if ( pos == cursor )
		{
			//TODO: -   .
			if ( current_time % 1000 > 500 ) RenderSprite(x, y, z, bbox['|'-32].Min.x, bbox['|'-32].Min.y, bbox['|'-32].Max.x, bbox['|'-32].Max.y, tex, false, tClr);
			x +=width['|'-32] + dist;
			RenderSprite(x, y, z, bbox[c].Min.x, bbox[c].Min.y, bbox[c].Max.x, bbox[c].Max.y, tex, false, tClr);
			x +=width[c] + dist;
		}
		else
		{
			RenderSprite(x, y, z, bbox[c].Min.x, bbox[c].Min.y, bbox[c].Max.x, bbox[c].Max.y, tex, false, tClr);

			x +=width[c] + dist;
		}
		pos++;
	}
	if ( pos == cursor && current_time % 1000 > 500 )
		RenderSprite(x, y, z, bbox['|'-32].Min.x, bbox['|'-32].Min.y, bbox['|'-32].Max.x, bbox['|'-32].Max.y, tex, false, tClr);
}

void TexFont::Print( const char *text, ... )
{
	if (text == NULL)
		return;

	//static UCHAR	*temp;//[FONT_STRING_ADD_LENGTH];
	//temp = new UCHAR [strlen(text)+FONT_STRING_ADD_LENGTH];
	SetBufLen(strlen(text) + FONT_STRING_ADD_LENGTH);
	va_list	ap;
	va_start(ap, text);
	vsnprintf((char*)buffer, buf_len, text, ap);
	va_end(ap);
	Draw(buffer);
	//delete [] temp;
}




UINT TexFont::GetStringWidth(const char *text)
{
	if (text == NULL)
		return 0;
	UINT r = 0;
	for (; *text; text++)
		r += (UINT)width[(unsigned char)(*text - 32)] + dist;
	return r;
}

UINT TexFont::GetStringWidth(const char *text, size_t t1, size_t t2)
{
	if (!text || t1 > t2)
		return 0;
	UINT r = 0;
	text = text + t1;
	for (;*text && t1 <= t2; text++, t1++)
		r += (UINT)width[(unsigned char)(*text - 32)] + dist;
	return r;
}

UINT TexFont::GetStringHeight(const char *text)
{
	if (text == NULL)
		return 0;
	UINT r = height[32];
	for (; *text; text++)
		r = max((UINT)height[(unsigned char)(*text - 32)], r);
	return r;
}

UINT TexFont::GetStringHeight(const char *text, size_t t1, size_t t2)
{
	if (!text || t1 > t2)
		return 0;
	UINT r = height[32];
	text = text + t1;
	for (;*text && t1 <= t2; text++, t1++)
		r = max((UINT)height[(unsigned char)(*text - 32)], r);
	return r;
}

UINT TexFont::GetWidth(char c)
{
	return (UINT)width[(unsigned char)(c-32)];
}
UINT TexFont::GetWidthAndDist(char c)
{
	return (UINT)width[(unsigned char)(c-32)] + dist;
}

UINT TexFont::GetHeight(char c)
{
	return (UINT)height[(unsigned char)(c-32)];
}

//            .
//     .
namespace ControlSeqParser
{
	//  :
	//		[0]
	//	'/'  |
	//		[1]
	//	'c'	/ \ 'n'
	//	  [2] [3]
	//
	const int states_count = 5;				//  .  states_count-1  - .
	const size_t classes_count = 4;			//   
	const CSeqType seqs[states_count] = {ENone, ENone, EColor, ENewLine, ENone};	//      
	const bool end_state[states_count] = {false, false, true, true, false};		//  
	const char classes[classes_count] = {'/', 'c', 'n', '\0'};					// 
	const int table[states_count][classes_count] = {	//   .  -  ,  - ,   -   .
		{1, 4, 4, 4},
		{4, 2, 3, 4},
		{4, 4, 4, 4},
		{4, 4, 4, 4},
		{4, 4, 4, 4} };
	
	const size_t seq_len[] = {0, 8, 2};		//       .   CSeqType.

	//    
	size_t GetClass(char c)
	{
		for (size_t i = 0; i < classes_count; i++)
		{
			if (classes[i] == c)
				return i;
		}
		return classes_count-1;
	}

	//  ,    ,        
	//  shift      .
	CSeqType CheckControlSeq(const char* text, size_t& shift)
	{
		shift = 0;
		int s = 0;
		size_t i;
		for (i = 0; text[i] != 0 && !end_state[s]; i++)
		{
			s = table[s][GetClass(text[i])];
			if (s == states_count - 1)
				break;
		}

		CSeqType t = seqs[s];
		if (t == ENone)
		{
			return ENone;
		}

		for (; text[i] != 0 && i < seq_len[t]; i++);

		if (i == seq_len[t])
		{
			shift = i;
			return t;
		}
		
		return ENone;
	}

	//        color  
	void GetColor(const char* text, RGBAf& color)
	{
		ASSERT(text[0] == '/' && text[1] == 'c');

		char number[3] = "\0\0";
		size_t shift = 2;
		for (int i = 0; i < 3; i++, shift+=2)
		{
			number[0] = text[shift];
			number[1] = text[shift+1];
			color[i] = strtoul(number, NULL, 16) / 255.0f;
		}
	}
};
