#ifndef __FONT_H_
#define __FONT_H_

#include "renderer.h"
#include "../misc.h"

//   
class IFont
{
public:
	Vector2 p;				//   
	float z;
	RGBAf tClr;				// .
	RGBAf outline;

	IFont()
	{
		z = 1.0f;
	}

	virtual ~IFont() { DELETEARRAY(buffer); buf_len = 0; }
	virtual bool Load() = 0;
	virtual void Recover() = 0;
	virtual void Print(const char *text, ...) = 0;
	//virtual void PrintMultiline(const char *text, CAABB &area);
	void PrintMultiline(const char *text, CAABB &area, int cursor = -1);

	virtual UINT GetStringWidth(const char *text) = 0;
	virtual UINT GetStringWidth(const char *text, size_t t1, size_t t2) = 0;
	virtual UINT GetStringHeight(const char *text) = 0;
	virtual UINT GetStringHeight(const char *text, size_t t1, size_t t2) = 0;

	virtual UINT GetWidth(char c) = 0;
	// TODO:  ,             .
	//     .
	virtual UINT GetWidthAndDist(char c) = 0;
	virtual UINT GetHeight(char c) = 0;
protected:
	virtual void Draw(const UCHAR *text) = 0;
	virtual void Draw(const UCHAR *text, int cursor) = 0;

	static UCHAR *buffer;		// -     .    ,    .
	static size_t buf_len;

	//    ,  ,  .
	static void SetBufLen(size_t len)
	{
		if (len <= buf_len) return;
		DELETEARRAY(buffer);
		//   ,    2.
		//  ,    ,       +1 .
		//          ,  ,         .
		buf_len = nextPowerOfTwo(len);
		buffer = new UCHAR[buf_len];
		buffer[0] = 0;
	}
};

//////////////////////////////////////////////////////////////////////////

bool LoadTextureFont(const char* in_name, const char* out_name);
void RecoverFonts();
//void FreeFont(WinFont* font);
void FreeFont(const char* font_name);
void FreeFonts(void);
IFont* FontByName(const char* font_name);

//Satana    

struct CRectf
{
public:
	Vector2 Min;
	Vector2 Max;
};


#define TEXFONT_DEFAULT_DISTANCE	1		//   
#define FONT_STRING_ADD_LENGTH	256			//  Print        ,   

class TexFont : public IFont
{
public:
	TexFont();
	~TexFont();

	string filename;	// ^W --   .

	//RGBAf				tClr;					//	.
	byte				dist;					//	  
	Vector2				wh;						//	     -.   . x - w, y - h
	bool				doUseGlobalCoordSystem;	//	      

	virtual bool Load();
	virtual void Recover();
	bool		SaveToFile();

	void		Print(const char *text, ...);
	//void		PrintMultiline(const char *text, CAABB &area);
	//void		PrintMultiline(const char *text, CAABB &area, int cursor);

	virtual UINT GetStringWidth(const char *text);
	virtual UINT GetStringWidth(const char *text, size_t t1, size_t t2);
	virtual UINT GetStringHeight(const char *text);
	virtual UINT GetStringHeight(const char *text, size_t t1, size_t t2);

	virtual UINT GetWidth(char c);
	virtual UINT GetWidthAndDist(char c);
	virtual UINT GetHeight(char c);
protected:
	CRectf		bbox[256];		//      
	byte		width[256];		//   
	byte		height[256];	//   
	char		*FontImageName;	//   
	const Texture* tex;

	virtual void Draw(const UCHAR *text);
	virtual void Draw(const UCHAR *text, int cursor);
};


namespace ControlSeqParser
{
	enum CSeqType {ENone, EColor, ENewLine};

	CSeqType CheckControlSeq(const char* text, size_t& shift);
	void GetColor(const char* text, RGBAf& color);
}

#endif // __FONT_H_
