#ifndef __RENDER_TYPES_H_
#define __RENDER_TYPES_H_

//  C4201: nameless struct/union,   RGBAf
#if _MSC_VER > 1000
#pragma warning (disable : 4201)
#endif

namespace Renderer
{

	union RGBAf
	{
	public:
		float v[4];
		struct {float x, y, z, w;};
		struct {float r, g, b, a;};

		__INLINE float operator[](int i)
		{
			if (i<0 || i> 3)
				return 0;
			return v[i];
		}


		RGBAf(){x=y=z=0.0f;w=1.0f;}
		RGBAf(float xv, float yv, float zv, float wv){x=xv;y=yv;z=zv;w=wv;}
		//void operator=(const RGBAf &q){x=q.x;y=q.y;z=q.z;w=q.w;}
		__INLINE RGBAf operator+(RGBAf q){return RGBAf( x + q.x , y + q.y, z + q.z, w + q.w );}
		__INLINE RGBAf operator-(RGBAf q){return RGBAf( x - q.x , y - q.y, z - q.z, w - q.w );}
		__INLINE RGBAf operator*(float s){return RGBAf(x * s, y * s, z * s, w * s);}
		__INLINE bool operator!=(RGBAf q){return x != q.x || y != q.y || z != q.z || w != q.w;}
		__INLINE RGBAf operator=(const RGBAf& V)
		{
			x = V.x;
			y = V.y;
			z = V.z;
			w = V.w;
			return *this;
		}


		__INLINE RGBAf operator/(float s)
		{
			if (s == 0.0f)
				return RGBAf();
			float is = 1.0f/s; 
			return (*this)*is;
		}
		__INLINE RGBAf operator+=(RGBAf q){(*this) = (*this)+q; return *this;}
		float Norm()const{return x * x + y * y + z * z + w * w;}
		float Length( void ){return sqrt(x * x + y * y + z * z + w * w );}	
		RGBAf Conjugate(){ return RGBAf(-x,-y,-z,w);}
		RGBAf Identity()
		{
			//TODO: Danger chek division by zero
			return (*this).Conjugate()/(*this).Norm();
		};
		float InnerProduct(const RGBAf& q){return x*q.x+y*q.y+z*q.z+w*q.w;}

		__INLINE RGBAf operator*(const RGBAf &q)
		{
			RGBAf result;

			result.x = w * q.x + x * q.w + y * q.z - z * q.y;
			result.y = w * q.y - x * q.z + y * q.w + z * q.x;
			result.z = w * q.z + x * q.y - y * q.x + z * q.w;
			result.w = w * q.w - x * q.x - y * q.y - z * q.z;

			return result;
		}

		__INLINE void glSet()
		{
			glColor4fv(&r);
		}
	};

	typedef RGBAf Quaternion;
}

struct  vertex2f_t{
	float x;
	float y;
	vertex2f_t():x(0),y(0){}
};

struct  vertex3f_t{
	float x;
	float y;
	float z;
	vertex3f_t():x(0),y(0),z(0){}
};

//struct  vertex3d_t{
//	double x;
//	double y;
//	double z;
//	vertex3d_t():x(0),y(0),z(0){}
//};


struct coord2f_t{
	float x;
	float y;
	coord2f_t():x(0),y(0){}
	coord2f_t(float x, float y):x(x),y(y){}
};


enum RenderSpriteMethod { rsmStandart, rsmStretch, rsmCrop, rsmRepeatX, rsmRepeatY, rsmRepeatXY };

#endif // __RENDER_TYPES_H_
